# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# curate.data_transform.py
"""Data Transformations"""

import logging
import re

import numpy as np
import pandas as pd
import patsy as pat

from ..utils.snakey_lowercaser import SnakeyLowercaser

__all__ = [
    "DatatypeConverter",
    "DatasetReshaper",
    "Transformer",
    "Standardizer",
    "compress_factor_levels",
]

_log = logging.getLogger(__name__)


class DatatypeConverter:
    """Force correct datatypes according to what model expects"""

    def __init__(self, ftsd: dict, date_format: str = "%Y-%m-%d"):
        """Initialise with fts and optionally specify factors with specific levels

        Use with a fts dict of form:
            ftsd = dict(
                fcat = [],           # use for unordered categoricals
                ford = {ft0: ['a', 'b', 'c'], ... },  # use for ordinals
                fstr = [],  # rarely used e.g. freetext strings / IDs
                fbool = [],
                fbool_nan_to_false = [],
                fdate = [],  # transform dates as str into datetime
                fyear = [],  # transform years as num/str into datetime
                fint = [],
                ffloat = [],
                fverbatim = [],        # maintain in current dtype)
        """
        self.ftsd = dict(
            fcat=ftsd.get("fcat", []),
            ford=ftsd.get("ford", {}),
            fstr=ftsd.get("fstr", []),
            fbool=ftsd.get("fbool", []),
            fbool_nan_to_false=ftsd.get("fbool_nan_to_false", []),
            fdate=ftsd.get("fdate", []),
            fyear=ftsd.get("fyear", []),
            fint=ftsd.get("fint", []),
            ffloat=ftsd.get("ffloat", []),
            fverbatim=ftsd.get("fverbatim", []),  # keep verbatim
        )
        self.rx_number_junk = re.compile(r"[#$€£₤¥,;%\s]")
        self.date_format = date_format
        inv_bool_dict = {
            True: ["yes", "y", "true", "t", "1", 1, 1.0],
            False: ["no", "n", "false", "f", "0", 0, 0.0],
        }
        self.bool_dict = {v: k for k, vs in inv_bool_dict.items() for v in vs}
        self.strnans = [
            "none",
            "nan",
            "null",
            "na",
            "n/a",
            "<na>",
            "missing",
            "empty",
            "",
        ]

    def convert_dtypes(self, dfraw: pd.DataFrame) -> pd.DataFrame:
        """Select fts and convert dtypes. Return cleaned df
        TODO keep up to date with pandas missing value handling
             https://pandas.pydata.org/docs/user_guide/missing_data.html
        """
        snl = SnakeyLowercaser()

        # subselect desired fts
        # TODO make this optional
        fts_all = [w for _, v in self.ftsd.items() for w in v]
        df = dfraw[fts_all].copy()

        for ft in self.ftsd["fcat"] + self.ftsd["fstr"]:
            # tame string, clean, handle nulls
            idx = df[ft].notnull()
            vals = df.loc[idx, ft].astype(str, errors="ignore").apply(snl.clean)
            df.drop(ft, axis=1, inplace=True)
            df.loc[~idx, ft] = pd.NA
            df.loc[idx, ft] = vals
            df.loc[df[ft].isin(self.strnans), ft] = pd.NA  # str reps of nulls
            if ft in self.ftsd["fcat"]:
                df[ft] = pd.Categorical(df[ft].values, ordered=False)

        for ft, lvls in self.ftsd["ford"].items():
            df[ft] = pd.Categorical(df[ft].values, categories=lvls, ordered=True)

        for ft in self.ftsd["fbool"] + self.ftsd["fbool_nan_to_false"]:
            # tame string, strip, lower, use self.bool_dict, use pd.NA
            if not isinstance(df.dtypes[ft], bool):
                # if df.dtypes[ft].type != bool:
                df[ft] = df[ft].apply(lambda x: str(x).strip().lower())
                df.loc[df[ft].isin(self.strnans), ft] = pd.NA
                df[ft] = df[ft].apply(lambda x: self.bool_dict.get(x, x))

                if ft in self.ftsd["fbool_nan_to_false"]:
                    df.loc[df[ft].isnull(), ft] = False

                set_tf_only = set(df[ft].unique())
                if len(set_tf_only - set([True, False])) == 0:  # most common
                    df[ft] = df[ft].astype(bool)
                elif pd.isnull(df[ft]).sum() > 0:  # contains NaNs, use pd.boolean
                    df[ft] = df[ft].convert_dtypes(convert_boolean=True)
                else:
                    # ft not yet properly mapped,
                    raise ValueError(
                        f"{ft} contains values ({set_tf_only}) incompatible with np.bool or pd.Boolean"
                    )

        for ft in self.ftsd["fyear"]:
            if isinstance(df.dtypes[ft], object):
                df[ft] = (
                    df[ft]
                    .astype(str)
                    .str.strip()
                    .str.lower()
                    .map(lambda x: self.rx_number_junk.sub(r"", x))
                )
                df.loc[df[ft].isin(self.strnans), ft] = np.nan
            df[ft] = df[ft].astype(float, errors="raise")
            if pd.isnull(df[ft]).sum() == 0:
                df[ft] = df[ft].astype(int, errors="raise")
            df[ft] = pd.to_datetime(df[ft], errors="raise", format="%Y")

        for ft in self.ftsd["fdate"]:
            df[ft] = pd.to_datetime(df[ft], errors="raise", format=self.date_format)

        for ft in self.ftsd["fint"]:
            try:
                if isinstance(df.dtypes[ft], object):
                    df[ft] = (
                        df[ft]
                        .astype(str)
                        .str.strip()
                        .str.lower()
                        .map(lambda x: self.rx_number_junk.sub(r"", x))
                    )
                    df.loc[df[ft].isin(self.strnans), ft] = np.nan
                df[ft] = df[ft].astype(float, errors="raise")
                if pd.isnull(df[ft]).sum() == 0:
                    df[ft] = df[ft].astype(int, errors="raise")
            except Exception as e:
                raise Exception(f"{str(e)} in ft: {ft}").with_traceback(
                    e.__traceback__
                ) from e

        for ft in self.ftsd["ffloat"]:
            try:
                if isinstance(df.dtypes[ft], object):
                    df[ft] = (
                        df[ft]
                        .astype(str)
                        .str.strip()
                        .str.lower()
                        .map(lambda x: self.rx_number_junk.sub(r"", x))
                    )
                    df.loc[df[ft].isin(self.strnans), ft] = np.nan
                df[ft] = df[ft].astype(float, errors="raise")
            except Exception as e:
                raise e(ft) from e

        for ft in self.ftsd["fverbatim"]:
            # just force missing to pd.NA
            idx = df[ft].isnull()
            df.loc[idx, ft] = pd.NA
            _log.info(f"Kept ft verbatim: {ft}")

        # NOTE we're now at the end of the columns

        return df[fts_all]


class DatasetReshaper:
    """Convenience functions to reshape whole datasets"""

    def __init__(self):
        pass

    def create_dfcmb(self, df: pd.DataFrame) -> pd.DataFrame:
        """Create a combination dataset `dfcmb` from inputted `df`, which
        MUST have correct / appropriate pandas datatypes.

        Datatypes must be understood by patsy: factor-values as
        pd.Categoricals, or bools, or ints or floats. No dates or strings

        Output `dfcmb` is NOT a big groupby (i.e cartesian join) of factor
        values. Instead, it's a compact concat of unique factor values
        (which can be horizontally ragged so we fill NaNs) and filler values
        for ints a floats

        We use `dfcmb` because as of the current patsy (0.5.1), design_info
        objects still can't be pickled, which means that when you want to
        transform a test / holdout sample according to a training dataset
        you have to instantiate a new design_info object in memory.

        In this codebase we set the design_info attribute on the
        Transformer object, so you can keep that object around in memory,
        but upon memory-loss you need to call it with the dfcmb created
        here e.g. Transformer.fit_transform(dfcmb).

        I recommend storing dfcmb in a DB: it only needs to be updated if
        the modelled features change (e.g. include a new feature) or if
        factor values (aka categorical levels) change
        """
        dfcmb = pd.DataFrame(index=[0])
        sdtypes = df.dtypes

        if (sum(sdtypes == "object") > 0) | (sum(sdtypes == "boolean") > 0):
            return (
                ValueError,
                "Valid dtypes are `category`, `bool`, `int`, `float` only",
            )
        cats = list(sdtypes.loc[sdtypes == "category"].index.values)
        bools = list(sdtypes.loc[sdtypes == "bool"].index.values)
        ints = list(sdtypes.loc[sdtypes == "int"].index.values)
        floats = list(sdtypes.loc[sdtypes == "float"].index.values)

        # create ragged cats
        for ft in cats:
            colnames_pre = list(dfcmb.columns.values)
            dfcmb = pd.concat(
                [dfcmb, pd.Series(df[ft].cat.categories)],
                axis=1,
                join="outer",
                ignore_index=True,
            )
            dfcmb.columns = colnames_pre + [ft]

        # apply categorical dtype again
        for ft in cats:
            dfcmb[ft] = pd.Categorical(
                dfcmb[ft].values,
                categories=df[ft].cat.categories,
                ordered=df[ft].cat.ordered,
            )

        for ft in bools:
            colnames_pre = list(dfcmb.columns.values)
            dfcmb = pd.concat(
                [dfcmb, pd.Series([False, True])],
                axis=1,
                join="outer",
                ignore_index=True,
            )
            dfcmb.columns = colnames_pre + [ft]

        for ft in bools:  # force to bool (we choose to not allow NaNs in bools)
            # dfcmb[ft] = dfcmb[ft].convert_dtypes(convert_boolean=True)
            dfcmb[ft] = dfcmb[ft].astype(bool)

        for ft in ints:
            dfcmb[ft] = 1

        for ft in floats:
            dfcmb[ft] = 1.0

        _log.info(
            f"Reduced df {len(df)} rows, {df.values.nbytes / 1e3:,.0f} KB)"
            + f" to dfcmb ({len(dfcmb)} rows, {dfcmb.values.nbytes / 1e3:,.0f} KB)"
        )

        return dfcmb


class Transformer:
    """Model-agnostic patsy transformer from row-wise natural observations
    into dmatrix according to patsy formula or patsy design_info object
    NOTES:
        + design_info is stateful
        + it's reasonable to initialise this per-observation but far more
          efficient to initialise once and persist in-memory
        + Categoricals must already be a pd.Categorical with all levels assigned
          appropriate to the full data domain. This will feed
          pd.Categorical.cat.codes (ints) representation into patsy
        + Booleans must already be np.bool (not pd.BooleanDtype, no NaNs)
    """

    def __init__(self):
        self.design_info = None
        self.factor_map = {}
        self.snl = SnakeyLowercaser()

    def _get_fts_to_force_to_int(self, dfraw: pd.DataFrame) -> list[str]:
        """Get list of ffts to force to int post patsy conversion"""
        s = dfraw.dtypes
        ints = list(s.loc[s == "int"].index.values)
        bools = [f"{ft}[T.True]" for ft in s.loc[s == "bool"].index.values]
        return ints + bools

    def _convert_cats(self, dfraw: pd.DataFrame) -> pd.DataFrame:
        """Common conversion of cats to codes and store mapping
        NOTE Missing values (np.nan) are recorded as cats.codes = -1
        which is helpfully still an int, so we can still return int dtype"""
        sdtypes = dfraw.dtypes
        df = dfraw.copy()
        cats = list(sdtypes.loc[sdtypes == "category"].index.values)
        for ft in cats:
            map_int_to_fct = dict(enumerate(df[ft].cat.categories))
            map_fct_to_int = {v: k for k, v in map_int_to_fct.items()}
            self.factor_map[ft] = map_fct_to_int
            df[ft] = df[ft].cat.codes.astype(int)
        return df

    def _transform(
        self,
        fml_or_design_info: str | pat.DesignInfo,
        df: pd.DataFrame,
        propagate_nans: bool,
    ) -> pd.DataFrame:
        """Common to fit_transform and transform"""
        # TODO add check for fml contains `~` and handle accordingly
        # TODO add option to output matrix   # np.asarray(mx_ex)
        df = self._convert_cats(df.copy())
        fts_force_to_int = self._get_fts_to_force_to_int(df)
        na_act = pat.NAAction(NA_types=[]) if propagate_nans else "raise"
        df_ex = pat.dmatrix(
            fml_or_design_info, df, NA_action=na_act, return_type="dataframe"
        )
        design_info = df_ex.design_info

        for ft in fts_force_to_int:
            if ft in df_ex.columns.values:
                df_ex[ft] = df_ex[ft].astype(int)

        return df_ex, design_info

    def fit_transform(
        self, fml: str, df: pd.DataFrame, propagate_nans: bool = True
    ) -> pd.DataFrame:
        """Fit a new `design_info` attribute for this instance according to
        `fml` acting upon `df`. Return transformed dmatrix as a DataFrame.
        Use this for a new training set or to initialise the transformer
        based on dfcmb.
        `fml` maps directly to feature names in `df` i.e. before patsy
        transforms and we take the dtypes from `df`.
        """
        df_ex, design_info = self._transform(fml, df, propagate_nans)
        self.design_info = design_info
        return df_ex

    def transform(self, df: pd.DataFrame, propagate_nans: bool = True) -> pd.DataFrame:
        """Transform input `df` to dmatrix according to pre-fitted
        `design_info`. Return transformed dmatrix (pd.DataFrame)
        """
        if self.design_info is None:
            raise AttributeError("No design_info, run `fit_transform()` first")

        df_ex, _ = self._transform(self.design_info, df, propagate_nans)
        return df_ex


class Standardizer:
    """Model-agnostic standardizer from pre-transformed dmatrix
    dmatrix according to patsy formula or patsy design_info object
    NOTES:
    + means, sdevs and scale are stateful: anticipate never changing
    + must be initialised with Transformer.design_info, consider refactoring
    + it's reasonable to initialise this per-observation but far more
        efficient to initialise once and persist in-memory
    + note these standardizations use nanmean, nanstd, so they're
        already compatible with recent changes (2021-03-31) to
        transformer to propagate nans
    + Automatically exclude post-patsy binary [T.x] features

    NEW FUNCTIONALITY:
    + apply standardization using a mask. allows us to exclude any col
    + rework to I/O dataframes

    TODO: introduce minmax scaling as an option
    """

    def __init__(self, tfmr: Transformer, fts_exclude: list = None):
        """Automatically exclude fts in list(tfmr.factor_map.keys()) and
        any named in `fts_exclude` that are numeric and would otherwise get
        standardardized"""

        if fts_exclude is None:
            fts_exclude = []
        self.design_info = tfmr.design_info
        self.fts_exclude = fts_exclude + list(tfmr.factor_map.keys())

        col_num_excl = [0] + [
            i
            for i, n in enumerate(self.design_info.column_names)
            if (n in self.fts_exclude) or re.search(r"\[T\.", n)
        ]

        # col_mask is True where we want to exclude the col from standardization
        self.col_mask = [
            True if i in col_num_excl else False
            for i in np.arange(len(self.design_info.column_names))
        ]

        self.means = None  # will become np.ndarray
        self.sdevs = None  # will become np.ndarray
        self.scale = None  # will become int

    def standardize(self, df: pd.DataFrame) -> pd.DataFrame:
        """Standardize input df to mean-centered, 2sd unit variance,
        Retain the fitted means and sdevs for later use in standardize()
        """
        if any([v is None for v in [self.means, self.sdevs, self.scale]]):
            raise AttributeError(
                "No mns, sdevs or scale, "
                + "run `fit_standardize()` on training set first"
            )

        df_s = (df - self.means) / (self.sdevs * self.scale)
        mask = np.tile(self.col_mask, (len(df), 1))

        # fill original df w/ standardized to more easily preseve dtype of ints
        df_exs = df.mask(~mask, df_s)  # replace values where condition is True
        return df_exs

    def fit_standardize(self, df: pd.DataFrame, scale: int = 2) -> pd.DataFrame:
        """Standardize numeric features of df with variable scale
        Retain the fitted means and sdevs for later use in standardize()
        """
        self.means = np.where(self.col_mask, np.nan, np.nanmean(df, axis=0))
        self.sdevs = np.where(self.col_mask, np.nan, np.nanstd(df, axis=0))
        self.scale = scale
        return self.standardize(df)

    def standardize_mx(self, mx: np.ndarray) -> np.ndarray:
        """Standardize input mx to mean-centered, 2sd unit variance,
        Retain the fitted means and sdevs for later use in standardize()
        """
        if any([v is None for v in [self.means, self.sdevs, self.scale]]):
            raise AttributeError(
                "No mns, sdevs or scale, "
                + "run `fit_standardize()` on training set first"
            )

        mxs_all = (mx - self.means) / (self.sdevs * self.scale)
        mask = np.tile(self.col_mask, (len(mx), 1))
        mxs = np.where(mask, mx, mxs_all)
        return mxs

    def fit_standardize_mx(self, mx: np.ndarray, scale: int = 2) -> np.ndarray:
        """Standardize numeric features of mx with variable scale
        Retain the fitted means and sdevs for later use in standardize()
        """
        # TODO add option to output dataframe
        self.means = np.where(self.col_mask, np.nan, np.nanmean(mx, axis=0))
        self.sdevs = np.where(self.col_mask, np.nan, np.nanstd(mx, axis=0))
        self.scale = scale
        return self.standardize(mx)

    def get_scale(self) -> tuple[pd.DataFrame, float]:
        """Get values followuing fit_standardize. Persist values over time."""
        means_sdevs = pd.DataFrame(
            {"means": self.means, "sdevs": self.sdevs},
            index=self.design_info.column_names,
        )

        return means_sdevs, self.scale

    def set_scale(self, df_means_sdevs: pd.DataFrame, scale: int = 2):
        """Set values saved from a prior fit_standardize. Now can run
        standardize for new data
        """
        self.means = df_means_sdevs["means"].values
        self.sdevs = df_means_sdevs["sdevs"].values
        self.scale = scale


def compress_factor_levels(df: pd.DataFrame, fts: list, topn: int = 20) -> pd.DataFrame:
    """Crude compression for factor levels, into the topn + 1 (other)
    Return new dataframe for fts
    """
    newdf = pd.DataFrame(index=df.index)
    for ft in fts:
        vc = df[ft].value_counts(dropna=False)
        _log.info(
            f"{ft}: compress {vc[:topn].sum()} ({vc[:topn].sum() / vc.sum():.1%})"
        )
        _log.info(
            f"{ft}: compressed {len(vc)}-{topn} ({len(vc) - topn}) levels "
            + "into `other`, "
            + f"{vc[topn:].sum()} rows ({vc[topn:].sum() / len(df):.1%}) affected"
        )
        vc_map = {
            k: (k if i < topn else "other")
            for i, (k, v) in enumerate(vc.to_dict().items())
        }
        newdf[f"{ft}_topn"] = df[ft].map(vc_map)

    return newdf



# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.plot.py
"""Model Plotting"""

from enum import Enum

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import xarray as xr
from matplotlib import figure, gridspec

from ..model_pymc import BasePYMCModel

__all__ = [
    "plot_trace",
    "plot_energy",
    "facetplot_krushke",
    "pairplot_corr",
    "forestplot_single",
    "forestplot_multiple",
    "plot_ppc",
    "plot_loo_pit",
    "plot_compare",
    "plot_lkjcc_corr",
]

sns.set_theme(
    style="darkgrid",
    palette="muted",
    context="notebook",
    rc={"figure.dpi": 72, "savefig.dpi": 144, "figure.figsize": (12, 4)},
)


class IDataGroupName(str, Enum):
    prior = "prior"
    posterior = "posterior"


def plot_trace(mdl: BasePYMCModel, rvs: list, **kwargs) -> figure.Figure:
    """Create traceplot for passed mdl NOTE a useful kwarg is `kind` e.g.
    'trace', the default is `kind = 'rank_vlines'`"""
    kind = kwargs.pop("kind", "rank_vlines")
    txtadd = kwargs.pop("txtadd", None)
    _ = az.plot_trace(
        mdl.idata, var_names=rvs, kind=kind, figsize=(12, 0.8 + 1.5 * len(rvs))
    )
    f = plt.gcf()
    _ = f.suptitle(
        " - ".join(filter(None, [f"Posterior traces of {rvs}", txtadd]))
        + f"\n{mdl.mdl_id}"
    )
    _ = f.tight_layout()
    return f


def plot_energy(mdl: BasePYMCModel) -> figure.Figure:
    """Simple wrapper around energy plot to provide a simpler interface"""
    _ = az.plot_energy(
        mdl.idata, fill_alpha=(0.8, 0.6), fill_color=("C0", "C8"), figsize=(12, 1.8)
    )
    f = plt.gcf()
    _ = f.suptitle(
        "NUTS Energy (Marginal vs Transitional, and E-BFMI)" + f" - `{mdl.mdl_id}`"
    )
    _ = f.tight_layout()
    return f


def facetplot_krushke(
    mdl: BasePYMCModel,
    rvs: list[str],
    group: IDataGroupName = IDataGroupName.posterior.value,
    m: int = 1,
    rvs_hack: int = 0,
    ref_vals: dict = None,
    **kwargs,
) -> figure.Figure:
    """Create Krushke-style plots using Arviz, univariate RVs, control faceting
    NOTE
    + ref_vals should look like a dict of list of dict with a key "ref_val"
        e.g. ref_vals = { 'beta_sigma' : [ {'ref_val':2} ] }
    + Optional Pass kwargs like hdi_prob = 0.5, coords = {'oid', oids}
    """
    # TODO unpack the compressed rvs from the idata
    txtadd = kwargs.pop("txtadd", None)
    transform = kwargs.pop("transform", None)
    n = 1 + ((len(rvs) + rvs_hack - m) // m) + ((len(rvs) + rvs_hack - m) % m)
    f, axs = plt.subplots(n, m, figsize=(3 * m, 0.8 + 1.5 * n))
    _ = az.plot_posterior(
        mdl.idata,
        group=group,
        ax=axs,
        var_names=rvs,
        ref_val=ref_vals,
        transform=transform,
        **kwargs,
    )
    _ = f.suptitle(
        " - ".join(filter(None, [f"Distribution of {rvs}", group, txtadd]))
        + f"\n{mdl.mdl_id}"
    )
    _ = f.tight_layout()
    return f


def forestplot_single(
    mdl: BasePYMCModel,
    rvs: list[str],
    group: IDataGroupName = IDataGroupName.posterior.value,
    **kwargs,
) -> figure.Figure:
    """Plot forestplot for list of rvs (optionally with factor sublevels)"""
    txtadd = kwargs.pop("txtadd", None)
    dp = kwargs.pop("dp", 2)
    plot_mn = kwargs.pop("plot_mn", True)
    transform = kwargs.pop("transform", None)
    desc = None
    kws = dict(
        colors=sns.color_palette("tab20c", n_colors=16).as_hex()[
            kwargs.pop("clr_offset", 0) :
        ][0],
        ess=False,
        combined=kwargs.pop("combined", True),
    )

    # get overall stats
    df = az.extract(mdl.idata, group=group, var_names=rvs).to_dataframe()
    if transform is not None:
        df = df.apply(transform)
    if len(rvs) == 1:
        mn = df[rvs[0]].mean(axis=0)
        qs = df[rvs[0]].quantile(q=[0.03, 0.97]).values
        desc = (
            f"Overall: $Mean =$ {mn:.{dp}f}"
            + ", $HDI_{94}$ = ["
            + ", ".join([f"{qs[v]:.{dp}f}" for v in range(2)])
            + "]"
        )
    nms = [nm for nm in df.index.names if nm not in ["chain", "draw"]]
    n = sum([len(df.index.get_level_values(nm).unique()) for nm in nms])

    f = plt.figure(figsize=(12, 1.5 + 0.2 * n))
    ax0 = f.add_subplot()
    _ = az.plot_forest(
        mdl.idata[group], var_names=rvs, **kws, transform=transform, ax=ax0
    )
    _ = ax0.set_title("")

    if plot_mn & (len(rvs) == 1):
        _ = ax0.axvline(mn, color="#ADD8E6", ls="--", lw=3, zorder=-1)
    else:
        _ = ax0.axvline(0, color="#ADD8E6", ls="--", lw=3, zorder=-1)
    _ = f.suptitle(
        "\n".join(
            filter(
                None,
                [
                    " - ".join(filter(None, [f"Forestplot of {rvs}", group, txtadd])),
                    mdl.mdl_id,
                    desc,
                ],
            )
        )
    )
    _ = f.tight_layout()
    return f


def forestplot_multiple(
    mdl: BasePYMCModel,
    datasets: dict[str, xr.core.dataarray.DataArray],
    group: IDataGroupName = IDataGroupName.posterior.value,
    **kwargs,
) -> figure.Figure:
    """Plot set of forestplots for related datasets RVs
    Useful for a linear model of RVs, where each RV can have sublevel factors
    TODO This makes a few too many assumptions, will improve in future
    """
    txtadd = kwargs.pop("txtadd", None)
    clr_offset = kwargs.pop("clr_offset", 0)
    dp = kwargs.pop("dp", 1)
    plot_med = kwargs.pop("plot_med", True)
    plot_combined = kwargs.pop("plot_combined", False)
    vsize = kwargs.pop("vsize", 10)
    desc = None

    hs = [0.22 * (np.prod(data.shape[2:])) for data in datasets.values()]
    f = plt.figure(figsize=(12, vsize + (0.5 * len(datasets)) + (0.2 * sum(hs))))
    gs = gridspec.GridSpec(len(hs), 1, height_ratios=hs, figure=f)

    for i, (txt, data) in enumerate(datasets.items()):
        ax = f.add_subplot(gs[i])
        _ = az.plot_forest(
            data,
            ax=ax,
            colors=sns.color_palette("tab20c", n_colors=16).as_hex()[clr_offset:][i],
            ess=False,
            combined=plot_combined,
        )

        _ = ax.set_title(txt)
        if plot_med:
            if i == 0:
                qs = np.quantile(data, q=[0.03, 0.25, 0.5, 0.75, 0.97])
                _ = ax.axvline(qs[2], color="#ADD8E6", ls="--", lw=3, zorder=-1)
                desc = (
                    f"med {qs[2]:.{dp}f}, HDI50 ["
                    + ", ".join([f"{qs[v]:.{dp}f}" for v in [1, 3]])
                    + "], HDI94 ["
                    + ", ".join([f"{qs[v]:.{dp}f}" for v in [0, 4]])
                    + "]"
                )
            else:
                _ = ax.axvline(1, color="#ADD8E6", ls="--", lw=3, zorder=-1)

    _ = f.suptitle(
        "\n".join(
            filter(
                None,
                [
                    " - ".join(filter(None, ["Forestplot levels", group, txtadd])),
                    mdl.mdl_id,
                    desc,
                ],
            )
        )
    )
    _ = f.tight_layout()
    return f


def pairplot_corr(
    mdl: BasePYMCModel,
    rvs: list[str],
    group: IDataGroupName = IDataGroupName.posterior.value,
    ref_vals: dict = None,
    **kwargs,
) -> figure.Figure:
    """Create posterior pair / correlation plots using Arviz, corrrlated RVs,
    Pass-through kwargs to az.plot_pair, e.g. ref_vals
    Default to posterior, allow for override to prior
    """
    txtadd = kwargs.pop("txtadd", None)
    kind = kwargs.pop("kind", "kde")

    pair_kws = dict(
        group=group,
        var_names=rvs,
        reference_values=ref_vals,  # deal with inconsistent naming
        divergences=True,
        marginals=True,
        kind=kind,
        kde_kwargs=dict(
            contourf_kwargs=dict(alpha=0.5, cmap="Blues"),
            contour_kwargs=dict(colors=None, cmap="Blues"),
            hdi_probs=[0.5, 0.94, 0.99],
        ),
        figsize=(2 + 1.8 * len(rvs), 2 + 1.8 * len(rvs)),
    )
    # idata[group][rvs].stack(dims=('chain', 'draw')).values.T,
    axs = az.plot_pair(mdl.idata, **pair_kws)
    corr = pd.DataFrame(
        az.sel_utils.xarray_to_ndarray(mdl.idata.get(group), var_names=rvs)[1].T
    ).corr()
    i, j = np.tril_indices(n=len(corr), k=-1)
    for ij in zip(i, j, strict=False):
        axs[ij].set_title(f"rho: {corr.iloc[ij]:.2f}", fontsize=8, loc="right", pad=2)
    vh_y = dict(rotation=0, va="center", ha="right")
    vh_x = dict(rotation=40, va="top", ha="right")
    _ = [a.set_ylabel(a.get_ylabel(), **vh_y) for ax in axs for a in ax]
    _ = [a.set_xlabel(a.get_xlabel(), **vh_x) for ax in axs for a in ax]

    f = plt.gcf()
    _ = f.suptitle(
        " - ".join(filter(None, ["Pairplot", mdl.name, group, "selected RVs", txtadd]))
    )
    _ = f.tight_layout()
    return f


def plot_ppc(
    mdl: BasePYMCModel,
    var_names: list,
    idata: az.InferenceData = None,
    group: str = "posterior",
    ecdf: bool = True,
    flatten: list = None,
    observed_rug: bool = True,
    logx: bool = False,
    **kwargs,
) -> figure.Figure:
    """Plot In- or Out-of-Sample Prior or Posterior Retrodictive. Does not
    require log-likelihood. Does require `observed_data`, which is not made by
    Potentials
    NOTE:
    + use var_names to only plot e.g. yhat
    + pass through kwargs, possibly of particular use is:
        `data_pairs` = {key (in observed_data): value (in {group}_predictive)}
        although we remind that the constant_data has the real name, but once
        it's observed in a log-likelihood the idata.observed_data will get the
        same name as the {group}_predictive, so data_pairs is not often needed
    """
    txtadd = kwargs.pop("txtadd", None)
    sharex = kwargs.pop("sharex", True)
    kind = "kde"
    kindnm = kind.upper()
    ynm = "density"
    loc = "upper right"
    n = len(var_names)
    if ecdf:
        kind = "cumulative"
        kindnm = "ECDF"
        ynm = "prop"
        loc = "lower right"

    if idata is None:
        _idata = mdl.idata
        insamp = True
    else:
        _idata = idata
        insamp = False

    if flatten is not None:
        n = 1
        for k in var_names:
            n *= _idata["observed_data"][k].shape[-1]
    # wild hack to get the size of observed
    i = list(dict(_idata.observed_data.sizes).values())[0]
    num_pp_samples = None if i < 500 else 200
    f, axs = plt.subplots(n, 1, figsize=(12, 1 + 2 * n), sharex=sharex, squeeze=False)
    _ = az.plot_ppc(
        _idata,
        group=group,
        kind=kind,
        ax=axs,
        var_names=var_names,
        flatten=flatten,
        observed_rug=observed_rug,
        random_seed=42,
        num_pp_samples=num_pp_samples,
        **kwargs,
    )
    _ = [ax.legend(fontsize=8, loc=loc) for ax in axs.flatten()]  # fix legend
    ls = None
    if logx:
        _ = [ax.set_xscale("log") for ax in axs.flatten()]
        ls = "(logscale)"
    _ = [
        ax.set(title=t, ylabel=ynm)
        for ax, t in zip(axs.flatten(), var_names, strict=False)
    ]
    t = f"{'In' if insamp else 'Out-of'}-sample {group.title()} Retrodictive {kindnm}"
    _ = f.suptitle(" - ".join(filter(None, [t, txtadd, ls])) + f"\n{mdl.mdl_id}")
    _ = f.tight_layout()
    return f


def plot_loo_pit(
    mdl: BasePYMCModel, data_pairs: dict = None, **kwargs
) -> figure.Figure:
    """Calc and plot LOO-PIT after run `mdl.sample_posterior_predictive()`
    ref: https://oriolabrilpla.cat/en/blog/posts/2019/loo-pit-tutorial.html
    NOTE:
    mdl.idata needs: observed_data AND log_likelihood AND posterior_predictive
    data_pairs {key (in observed AND log_likelihood): value (in posterior_predictive)}

    """
    txtadd = kwargs.pop("txtadd", None)
    f, axs = plt.subplots(
        len(data_pairs), 2, figsize=(12, 3 * len(data_pairs)), squeeze=False
    )
    for i, (y, yhat) in enumerate(data_pairs.items()):
        kws = dict(y=y, y_hat=yhat)
        _ = az.plot_loo_pit(mdl.idata, **kws, ax=axs[i][0], **kwargs)
        _ = az.plot_loo_pit(mdl.idata, **kws, ax=axs[i][1], ecdf=True, **kwargs)

        _ = axs[i][0].set_title(f"Predicted {yhat} LOO-PIT")
        _ = axs[i][1].set_title(f"Predicted {yhat} LOO-PIT cumulative")

    _ = f.suptitle(
        " - ".join(filter(None, ["In-sample LOO-PIT", txtadd])) + f"\n{mdl.mdl_id}"
    )
    _ = f.tight_layout()
    return f


def plot_compare(
    mdl_dict: dict[str, BasePYMCModel], yhats: list[str], **kwargs
) -> tuple[figure.Figure, dict[str, pd.DataFrame]]:
    """Calc and plot model comparison in-sample via expected log pointwise
    predictive density (ELPD) using LOO
    NOTE:
    idata needs: observed_data AND log_likelihood
    hats should be the key for observed_data AND log_likelihood
    """
    txtadd = kwargs.pop("txtadd", None)
    sharex = kwargs.pop("sharex", False)
    f, axs = plt.subplots(
        len(yhats),
        1,
        figsize=(12, 2.6 * len(yhats) + 0.2 * len(mdl_dict)),
        squeeze=False,
        sharex=sharex,
    )
    # mdlnms = ' vs '.join(idata_dict.keys())
    idata_dict = {f"{k}\n{v.mdl_id_fn}": v.idata for k, v in mdl_dict.items()}
    dcomp = {}
    for i, y in enumerate(yhats):
        dfcomp = az.compare(
            idata_dict, var_name=y, ic="loo", method="stacking", scale="log"
        )
        dcomp[y] = dfcomp
        ax = az.plot_compare(
            dfcomp, ax=axs[i][0], title=False, textsize=10, legend=False
        )
        _ = ax.set_title(y)
    t = (
        "Model Performance Comparison: ELPD via In-Sample LOO-PIT:\n`"
        + "` vs `".join(list(mdl_dict.keys()))
        + "`\n(higher & narrower is better)"
    )
    _ = f.suptitle(" - ".join(filter(None, [t, txtadd])))
    _ = f.tight_layout()

    return f, dcomp


def plot_lkjcc_corr(mdl: BasePYMCModel, **kwargs) -> figure.Figure:
    """Plot lkjcc_corr model RVs
    Drop diagonals, assume coord is called lkjcc_corr
    Also see https://python.arviz.org/en/stable/user_guide/label_guide.html#custom-labellers
    """
    coords = {
        "lkjcc_corr_dim_0": xr.DataArray([0, 1], dims=["asdf"]),
        "lkjcc_corr_dim_1": xr.DataArray([1, 0], dims=["asdf"]),
    }

    return facetplot_krushke(
        mdl=mdl,
        txtadd="lkjcc_corr, diagonals only",
        rvs=["lkjcc_corr"],
        coords=coords,
        m=2,
        rvs_hack=0,
        **kwargs,
    )


def plot_yhat_vs_y(
    mdl: BasePYMCModel,
    dfhat: pd.DataFrame,
    yhat: str = "yhat",
    y: str = "y",
    oid: str = "oid",
    insamp: bool = False,
    **kwargs,
) -> figure.Figure:
    """Boxplot forecast yhat with overplotted y"""
    txtadd = kwargs.pop("txtadd", None)
    kws_mn = dict(
        markerfacecolor="w", markeredgecolor="#333333", marker="d", markersize=12
    )
    kws_box = dict(
        kind="box", showfliers=False, showmeans=True, whis=(3, 97), meanprops=kws_mn
    )
    kws_sctr = dict(s=80, color="#32CD32")

    g = sns.catplot(
        x=yhat, y=oid, data=dfhat.reset_index(), **kws_box, height=4, aspect=3
    )
    _ = g.map(sns.scatterplot, y, oid, **kws_sctr, zorder=100)
    t_io = (
        f"{'In' if insamp else 'Out-of'}-sample: boxplots of posterior `{yhat}`"
        + f" with overplotted actual `{y}` values per observation"
        + f" `{oid}` (green dots) - `{mdl.name}`"
    )
    _ = g.fig.suptitle(" - ".join(filter(None, [t_io, txtadd])) + f"\n{mdl.mdl_id}")

    _ = g.tight_layout()
    return g.fig



# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.calc.py
"""Common Calculations for Model Evaluation"""

import sys
from typing import Any, Dict, Optional, Sequence, Tuple, cast

import numpy as np
import pandas as pd
import pymc
import pytensor
import pytensor.gradient as tg
import pytensor.tensor as pt
from arviz import InferenceData, dict_to_dataset

# from fastprogress import progress_bar
from pymc.backends.arviz import _DefaultTrace, coords_and_dims_for_inferencedata
from pymc.model import Model, modelcontext
from pymc.pytensorf import PointFunc

__all__ = [
    "get_log_jcd_scalar",
    "get_log_jcd_scan",
    "calc_f_beta",
    "calc_binary_performance_measures",
    "calc_rmse",
    "calc_r2",
    "calc_bayesian_r2",
    "calc_ppc_coverage",
    "expand_packed_triangular",
    "calc_2_sample_delta_prop",
    "numpy_invlogit",
    # "compute_log_likelihood_for_potential",
]


RSD = 42
rng = np.random.default_rng(seed=RSD)


def get_log_jcd_scalar(
    f_inv_x: pt.TensorVariable, x: pt.TensorVariable
) -> pt.TensorVariable:
    """Calc log of determinant of Jacobian where f_inv_x and x are both (n,)
    dimensional tensors, and `f_inv_x` is a direct, element-wise transformation
    of `x` without cross-terms (diagonals). Add this Jacobian adjustment to
    models where observed is a transformation, to handle change in coords / volume.
    NOTE:
    + Slight abuse of terminology because for a 1D f_inv_x, we can sum to scalar
    + Initially developed for a model with 1D f_inv_x

    Also see detail from Stan docs:
    + https://archive.ph/C9hWB
    + https://mc-stan.org/documentation/case-studies/mle-params.html
    + https://mc-stan.org/docs/2_25/reference-manual/change-of-variables-section.html#multivariate-changes-of-variables
    "The absolute derivative of the inverse transform measures how the scale of
    the transformed variable changes with respect to the underlying variable."

    Example from jungpenglao:
    + https://github.com/junpenglao/Planet_Sakaar_Data_Science/blob/4366de036cc608c942fdebb930e96f2cc8b83d71/Ports/Jacobian%20Adjustment%20in%20PyMC3.ipynb#L163  # noqa: W505
    + https://github.com/junpenglao/Planet_Sakaar_Data_Science/blob/main/WIP/vector_transformation_copulas.ipynb

    More discussion:
    + https://www.tamaspapp.eu/post/jacobian-chain/
    + https://discourse.pymc.io/t/jacobian-adjustment/1711
    + https://github.com/junpenglao/All-that-likelihood-with-PyMC3/blob/master/Notebooks/Neals_funnel.ipynb
    + https://discourse.pymc.io/t/how-do-i-implement-an-upper-limit-log-normal-distribution/1337/4
    + https://github.com/junpenglao/advance-bayesian-modelling-with-PyMC3/blob/master/Advance_topics/Box-Cox%20transformation.ipynb  # noqa: W505
    + https://slideslive.com/38907842/session-3-model-parameterization-and-coordinate-system-neals-funnel
    + https://discourse.pymc.io/t/mixture-model-with-boxcox-transformation/988
    + https://pytensor.readthedocs.io/en/latest/library/gradient.html#pytensor.gradient.grad
    + https://www.pymc.io/projects/docs/en/latest/_modules/pymc/math.html#
    + https://www.cs.toronto.edu/~rgrosse/courses/csc321_2018/slides/lec10.pdf
    + https://github.com/pymc-devs/pymc-examples/blob/1428f1b4e0d352a88667776b3ec612db93e032d9/examples/case_studies/copula-estimation.ipynb  # noqa: W505
    + https://discourse.pymc.io/t/when-to-adjust-for-jacobian-and-when-to-skip-in-pymc3/2830
    """
    return pt.log(pt.abs(tg.grad(cost=pt.sum(f_inv_x), wrt=[x])))


def get_log_jcd_scan(
    f_inv_x: pt.TensorVariable,
    x: pt.TensorVariable,
    upstream_rvs: list[pt.TensorVariable] = None,
) -> pt.TensorVariable:
    """Calc log of determinant of Jacobian where f_inv_x and x are both (n, k)
    dimensional tensors, and `f_inv_x` is a direct, element-wise transformation
    of `x` without cross-terms (diagonals). Add this Jacobian adjustment to
    models where observed is a transformation, to handle change in coords / volume.
    NOTE:
    + Initially developed for a model where k = 2
    + Use explicit scan and strict, make sure to pass in any RVs upstream of
      f_inv_x in `upstream_rvs`: scan needs to see these in non_sequences,
      although we dont actually use them inside the inner function get_grads
    + Break into two scan calls to handle each k separately
    + Scales linearly with n, so expect to take a while for n > 200
    + Based on https://github.com/pymc-devs/pytensor/blob/7bb18f3a3590d47132245b7868b3a4a6587a4667/pytensor/gradient.py#L1984  # noqa W505
    + Also see https://discourse.pymc.io/t/something-changed-in-pytensor-2-12-3-and-thus-pymc-5-6-1-that-makes-my-pytensor-gradient-grad-call-get-stuck-any-ideas/13100/4  # noqa W505
    + Also see https://discourse.pymc.io/t/hitting-a-weird-error-to-do-with-rngs-in-scan-in-a-custom-function-inside-a-potential/13151/14  # noqa W505
    """
    n = f_inv_x.shape[0]
    idx = pt.arange(n, dtype="int32")
    non_seq = [f_inv_x, x]
    if upstream_rvs is not None:
        non_seq += upstream_rvs

    def _grads(i, s, c, w, *args):
        """Inner function allows for extra args that we dont actually use:
        because scan needs to know about upstream_rvs in non_sequences and also
        wants to passes them into the inner function `_grad`.
        See https://github.com/pymc-devs/pytensor/pull/191

        """
        return tg.grad(cost=c[i, s], wrt=[w])  # shape (n, n, k)

    kws = dict(fn=_grads, sequences=idx, n_steps=n, name="_grads", strict=True)
    grads0, _ = pytensor.scan(**kws, non_sequences=[0, *non_seq])
    grads1, _ = pytensor.scan(**kws, non_sequences=[1, *non_seq])
    grads = grads0.sum(axis=0) + grads1.sum(axis=0)  # shape (n, k)
    log_jcd = pt.sum(pt.log(pt.abs(grads)), axis=1)  # shape (n)
    return log_jcd


def calc_f_beta(precision: np.array, recall: np.array, beta: float = 1.0) -> np.array:
    """Set beta such that recall is beta times more important than precision"""
    with np.errstate(divide="ignore", invalid="ignore"):
        fb = (1 + beta**2) * (precision * recall) / ((beta**2 * precision) + recall)
    return np.nan_to_num(fb, nan=0, posinf=0, neginf=0)


def calc_binary_performance_measures(y: np.array, yhat: np.array) -> pd.DataFrame:
    """Calculate tpr (recall), fpr, precision, accuracy for binary target,
    using quantiles of all samples from PPC, use vectorised calcs
    shapes y: (nsamples,), yhat: (nsamples, nobservations)
    """
    qs = np.round(np.arange(0, 1.01, 0.01), 2)
    yhat_q = np.quantile(yhat, qs, axis=0, method="linear").T
    y_mx = np.tile(y.reshape(-1, 1), 101)

    # calc tp, fp, tn, fn vectorized
    tp = np.nansum(np.where(yhat_q == 1, y_mx, np.nan), axis=0)
    fp = np.nansum(np.where(yhat_q == 1, 1 - y_mx, np.nan), axis=0)
    tn = np.nansum(np.where(yhat_q == 0, 1 - y_mx, np.nan), axis=0)
    fn = np.nansum(np.where(yhat_q == 0, y_mx, np.nan), axis=0)

    # calc tpr (recall), fpr, precision etc
    with np.errstate(divide="ignore", invalid="ignore"):
        accuracy = (tp + tn) / (tp + tn + fp + fn)
        tpr = recall = tp / (tp + fn)
        fpr = fp / (tn + fp)
        precision = np.nan_to_num(tp / (tp + fp), nan=1)  # beware of divide by zero

    perf = pd.DataFrame(
        {
            "accuracy": accuracy,
            "tpr": tpr,
            "fpr": fpr,
            "recall": recall,
            "precision": precision,
            "f0.5": calc_f_beta(precision, recall, beta=0.5),
            "f1": calc_f_beta(precision, recall, beta=1),
            "f2": calc_f_beta(precision, recall, beta=2),
        },
        index=qs,
    )
    perf.index.set_names("q", inplace=True)

    return perf.round(6)


def calc_rmse(
    dfhat: pd.DataFrame,
    oid: str = "oid",
    yhat: str = "yhat",
    y: str = "y",
    method: str = "a",
    qs: bool = False,
    mse_only: bool = False,
) -> float | tuple[float, pd.Series]:
    r"""Calculate (R)MSE using the mean PPC samples:

    Require that dfhat is a long-format left join of PPC samples `yhat` and
    observed `y` values (which of course will contain nsamples duplicates)
    ie.: len(dfhat) == (nobs * nsamples)
    and: dfhat has index [oid, chain, draw] and fts [yhat, y]

    Optionally set mse_only = True to get just the MSE (default is RMSE)
    Optionally set qs = True to also return a Series of quantiles

    For the SE calc, we have two options:
    + A: compress samples to a summary statistic before calculate MSE
    + B: calculate MSE for all samples and then take summary statistic

    \begin{align}
    \text{A:} \text{MSE} &= \frac{1}{n} \sum_{i=1}^{i=n}(\mu_{j}(\hat{y}_{i}) - y_{i})^{2} \\
    \text{B:} \text{MSE} &= \frac{1}{n} \sum_{i=1}^{i=n}(\mu_{j}(\hat{y}_{ij} - y_{i})^{2})
    \end{align}

    Option B can create huge MSE values because outliers for each obs are
    squared and push farther from the mean. This might be useful for a very
    sensitive analysis.

    i.e.
    take mean across samples then square the differences is usually smaller
    than square the differences each sample and preserve samples
    https://en.wikipedia.org/wiki/Generalized_mean

    But here we will default to the easier-to-understand Option A
    """
    if method not in ["a", "b"]:
        raise AttributeError("method must be in `a` or `b`")

    def _grp_se_at_mn(g):
        if method == "a":
            return np.power(g[yhat].mean() - g[y].max(), 2)
        else:
            return np.power(g[yhat] - g[y], 2).mean()

    def _grp_se_at_qs(g):
        qs = np.round(np.linspace(0, 1, 100 + 1), 2)
        if method == "a":
            se_at_qs = np.power(np.quantile(g[yhat], qs) - g[y].max(), 2)
        else:
            se_at_qs = np.quantile(np.power(g[yhat] - g[y], 2), qs)

        s_se_at_qs = pd.Series(se_at_qs, index=qs, name="se")
        s_se_at_qs.index.rename("q", inplace=True)
        return s_se_at_qs

    mse_at_mn = dfhat.groupby(level=oid).apply(_grp_se_at_mn).mean(axis=0)
    if not qs:
        if mse_only:
            return mse_at_mn
        else:
            return np.sqrt(mse_at_mn)
    else:
        mse_at_qs = dfhat.groupby(level=oid).apply(_grp_se_at_qs).mean(axis=0)
        if mse_only:
            return mse_at_mn, mse_at_qs
        else:
            rmse_at_qs = mse_at_qs.map(np.sqrt)
            rmse_at_qs._set_name("rmse", inplace=True)
            return np.sqrt(mse_at_mn), rmse_at_qs


def calc_r2(y: np.ndarray, yhat: np.ndarray) -> tuple[np.ndarray, pd.Series]:
    """Calculate R2 at mean and quantiles (summary stats)
    shapes: y (nobs), yhat (nobs, nsamples)
    $$R^{2} = 1 - \\frac{\\sum e_{model}^{2}}{\\sum e_{mean}^{2}}$$
    R2 normal range [0, 1]
    """
    sse_mean = np.sum((y - y.mean(axis=0)) ** 2)

    # Collapse samples to mean then calc error
    sse_model_mean = np.sum((y - yhat.mean(axis=1)) ** 2)
    r2_mean = 1 - (sse_model_mean / sse_mean)

    # calc summary stats of yhat
    smry = np.arange(0, 101, 5)
    sse_model = np.sum(
        (y - np.percentile(yhat, smry, axis=1)) ** 2, axis=1
    )  # (len(smry), nobs)
    r2_pct = pd.Series(1 - (sse_model / sse_mean), index=smry, name="r2")
    r2_pct.index.rename("pct", inplace=True)

    return r2_mean, r2_pct


def calc_bayesian_r2(y: np.ndarray, yhat: np.ndarray) -> pd.DataFrame:
    """Calculate a much better Bayesian R2 using all samples
    shapes: y (nobs,), yhat (nobs, nsamples), return: (nsamples, )
    Related to arviz.r2_score
    """
    var_yhat = np.var(yhat, axis=0)
    var_residuals = np.var(y.reshape(-1, 1) - yhat, axis=0)
    r2 = var_yhat / (var_yhat + var_residuals)
    return pd.DataFrame(r2, columns=["r2"])


def calc_ppc_coverage(y: np.ndarray, yhat: np.ndarray) -> pd.DataFrame:
    """Calc the proportion of coverage from full yhat ppc
    shapes: y (nobservations), yhat (nobservations, nsamples)
    """

    crs = np.arange(0, 1.01, 0.02)
    bounds = dict(
        pin_left=dict(
            lower=np.tile(np.percentile(yhat, 0.0, axis=1), reps=(len(crs), 1)),
            upper=np.percentile(yhat, 100.0 * crs, axis=1),
        ),
        middle_out=dict(
            lower=np.percentile(yhat, 50.0 - (50.0 * crs), axis=1),
            upper=np.percentile(yhat, 50.0 + (50.0 * crs), axis=1),
        ),
        # pin_right=dict(       ##just a rotation of pin_left
        #     lower=np.percentile(yhat, 100. - (100 * crs), axis=0),
        #     upper=np.tile(np.percentile(yhat, 100., axis=0), reps=(len(crs), 1)))
    )
    cov = []
    for k, v in bounds.items():
        for i, cr in enumerate(crs):
            cov.append(
                (
                    k,
                    cr,
                    np.sum(np.int64(y >= v["lower"][i]) * np.int64(y <= v["upper"][i]))
                    / len(y),
                )
            )

    return pd.DataFrame(cov, columns=["method", "cr", "coverage"])


def expand_packed_triangular(n, packed, lower=True, diagonal_only=False):
    r"""Convert a packed triangular matrix into a two dimensional array.
    Triangular matrices can be stored with better space efficiancy by
    storing the non-zero values in a one-dimensional array. We number
    the elements by row like this (for lower or upper triangular matrices):
        [[0 - - -]     [[0 1 2 3]
         [1 2 - -]      [- 4 5 6]
         [3 4 5 -]      [- - 7 8]
         [6 7 8 9]]     [- - - 9]
    Parameters
    ----------
    n: int
        The number of rows of the triangular matrix.
    packed: theano.vector
        The matrix in packed format.
    lower: bool, default=True
        If true, assume that the matrix is lower triangular.
    diagonal_only: bool
        If true, return only the diagonal of the matrix.
    """
    if packed.ndim != 1:
        raise ValueError("Packed triagular is not one dimensional.")
    if not isinstance(n, int):
        raise TypeError("n must be an integer")

    if diagonal_only and lower:
        diag_idxs = np.arange(1, n + 1).cumsum() - 1
        return packed[diag_idxs]
    elif diagonal_only and not lower:
        diag_idxs = np.arange(2, n + 2)[::-1].cumsum() - n - 1
        return packed[diag_idxs]
    elif lower:
        out = pt.zeros((n, n), dtype=pytensor.config.floatX)
        idxs = np.tril_indices(n)
        return pt.set_subtensor(out[idxs], packed)
    elif not lower:
        out = pt.zeros((n, n), dtype=pytensor.config.floatX)
        idxs = np.triu_indices(n)
        return pt.set_subtensor(out[idxs], packed)


def calc_2_sample_delta_prop(a, aref, a_index=None, fully_vectorised=False):
    r"""Calculate 2-side sample delta difference between arrays row-wise
    so that we can make a statement about the difference between a test array
    a reference array how different a is from aref

    Basic algo
    ----------

    for each row i in a:
        for each row j in aref:
            do: d = a[i] - aref[j]
            do: q = quantiles[0.03, 0.97](d)
            do: b_i = q > 0
            if: sum(b_i) == 2:
                we state "97% of a[i] > aref[j], substantially larger"
            elif: sum(b_i) == 1:
                we state "not different"
            else (sum(b_i) == 0):
                we state "97% of a[i] < aref[j], substantially smaller"
        do: prop = unique_count(b) / len(b)

        we state "prop a[i] larger | no different | smaller than aref"

    Parameters
    ----------
    a: 2D numpy array shape (nobs, nsamples), as returned by sample_ppc()
        This will be tested against the reference array arr_ref
        This is typically the prediction set (1 or more policies)

    aref: 2D numpy array shape (nobs, nsamples), as returned by sample_ppc()
        This is typically the training set (1 or more policies)

    a_index: Pandas series index, default None
        If a_index is not None, then prop_delta is returned as a DataFrame

    fully_vectorised : bool, default=False
        arr is not limited to a single policy
        If True, we use a numpy broadcasting to create a 3D array of deltas
            and perform both loops i, j in vectorised fashion
            see https://stackoverflow.com/a/43412438
            This is clever but consumes a lot of RAM (GBs)
        If False we loop the outer loop i
            This is more memory efficient and in testing seems faster!

    Returns
    -------
    prop_delta : numpy array of proportions of arr that are
        [0, 1, 2](substantially lower, no difference, substantially higher)
        than aref
        has shape (len(a), 3)
    """

    def _bincount_pad(a, maxval=2):
        b = np.bincount(a)
        return np.pad(
            b, (0, np.maximum(0, maxval + 1 - len(b))), "constant", constant_values=(0)
        )

    # silently deal with the common mistake of sending a 1D array for testing
    # must be a horizontal slice
    if a.ndim == 1:
        a = a[np.newaxis, :]

    rope = np.array([0.03, 0.97])  # ROPE limits, must be len 2

    if fully_vectorised:
        delta = a[:, np.newaxis] - aref  # (len(a), len(aref), width(aref))
        delta_gt0 = 1 * (
            np.quantile(delta, rope, axis=2) > 0
        )  # (len(rope), len(a), len(aref))
        n_intersects = np.sum(delta_gt0, axis=0)  # (len(a), len(aref))

    else:
        n_intersects = np.empty(shape=(len(a), len(aref)))  # (len(a), len(aref))
        for i in range(len(a)):
            delta = a[i] - aref  # (len(aref), width(aref))
            delta_gt0 = 1 * (
                np.quantile(delta, rope, axis=1) > 0
            )  # (len(rope), len(aref))
            n_intersects[i] = np.sum(delta_gt0, axis=0)  # (len(aref))
        n_intersects = n_intersects.astype(np.int)

    prop_intersects_across_aref = np.apply_along_axis(
        lambda r: _bincount_pad(r, len(rope)), 1, n_intersects
    ) / len(aref)  # (len(a), [0, 1, 2])

    if a_index is not None:
        prop_intersects_across_aref = pd.DataFrame(
            prop_intersects_across_aref,
            columns=["subs_lower", "no_difference", "subs_higher"],
            index=a_index,
        )

    return prop_intersects_across_aref


def numpy_invlogit(x, eps=sys.float_info.epsilon):
    """The inverse of the logit function, 1 / (1 + exp(-x))."""
    return (1.0 - 2.0 * eps) / (1.0 + np.exp(-x)) + eps


# def compute_log_likelihood_for_potential(
#     idata: InferenceData,
#     *,
#     var_names: Optional[Sequence[str]] = None,
#     extend_inferencedata: bool = True,
#     model: Optional[Model] = None,
#     sample_dims: Sequence[str] = ("chain", "draw"),
#     progressbar=True,
# ):
#     """Hackidy hack JS @ 20230919
#     Copied and modified from pymc compute_log_likelihood to allow a Potential
#     Differences NOTED with inline comments
#     orig: https://github.com/pymc-devs/pymc/blob/92278278d4a8b78f17ed0f101eb29d0d9982eb45/pymc/stats/log_likelihood.py#L29C1-L128C31
#     discussion: https://discourse.pymc.io/t/using-a-random-variable-as-observed/7184/10

#     IMPORTANT NOTE 2024-08-04 in the intervening time, the source function that
#     this copies / modifies has changed hugely - it's going to cause substantial
#     pain to update :S

# IMPORTANTER NOTE 2026-01-18 dataset_to_point_list no longer available
# in pymc v5.27, so remove this function for now, mayneed to return as/when
# we need this

#     ---

#     Compute elemwise log_likelihood of model given InferenceData with posterior group

#     Parameters
#     ----------
#     idata : InferenceData
#         InferenceData with posterior group
#     var_names : sequence of str, optional
#         List of Potential variable names for which to compute log_likelihood
#     extend_inferencedata : bool, default True
#         Whether to extend the original InferenceData or return a new one
#     model : Model, optional
#     sample_dims : sequence of str, default ("chain", "draw")
#     progressbar : bool, default True

#     Returns
#     -------
#     idata : InferenceData
#         InferenceData with log_likelihood group

#     """

#     posterior = idata["posterior"]

#     model = modelcontext(model)

#     observed_vars = [model.named_vars[name] for name in var_names]

#     if var_names is None:
#         observed_vars = model.observed_RVs
#         var_names = tuple(rv.name for rv in observed_vars)
#     else:
#         observed_vars = [model.named_vars[name] for name in var_names]
#         if not set(observed_vars).issubset(
#             model.observed_RVs + model.potentials
#         ):  # NOTE MODIFIED JS
#             raise ValueError(
#                 f"var_names must refer to observed_RVs in the model. Got: {var_names}"
#             )

#     # We need to temporarily disable transforms, because the InferenceData only keeps the untransformed values
#     # pylint: disable=used-before-assignment
#     try:
#         original_rvs_to_values = model.rvs_to_values
#         original_rvs_to_transforms = model.rvs_to_transforms

#         model.rvs_to_values = {
#             rv: rv.clone() if rv not in model.observed_RVs else value
#             for rv, value in model.rvs_to_values.items()
#         }
#         model.rvs_to_transforms = dict.fromkeys(model.basic_RVs)

#         elemwise_loglike_fn = model.compile_fn(
#             inputs=model.value_vars,
#             outs=model.logp(vars=observed_vars, sum=False),
#             on_unused_input="ignore",
#         )
#         elemwise_loglike_fn = cast(PointFunc, elemwise_loglike_fn)
#     finally:
#         model.rvs_to_values = original_rvs_to_values
#         model.rvs_to_transforms = original_rvs_to_transforms
#     # pylint: enable=used-before-assignment

#     # Ignore Deterministics
#     posterior_values = posterior[[rv.name for rv in model.free_RVs]]
#     posterior_pts, stacked_dims = dataset_to_point_list(posterior_values, sample_dims)
#     n_pts = len(posterior_pts)
#     loglike_dict = _DefaultTrace(n_pts)
#     indices = range(n_pts)
#     # if progressbar:
#     #     indices = progress_bar(indices, total=n_pts, display=progressbar)

#     for idx in indices:
#         loglikes_pts = elemwise_loglike_fn(posterior_pts[idx])
#         for rv_name, rv_loglike in zip(var_names, loglikes_pts, strict=False):
#             loglike_dict.insert(rv_name, rv_loglike, idx)

#     loglike_trace = loglike_dict.trace_dict
#     for key, array in loglike_trace.items():
#         loglike_trace[key] = array.reshape(
#             (*[len(coord) for coord in stacked_dims.values()], *array.shape[1:])
#         )

#     coords, dims = coords_and_dims_for_inferencedata(model)
#     loglike_dataset = dict_to_dataset(
#         loglike_trace,
#         library=pymc,
#         dims=dims,
#         coords=coords,
#         default_dims=list(sample_dims),
#         skip_event_dims=True,
#     )

#     if extend_inferencedata:
#         idata.add_groups(dict(log_likelihood=loglike_dataset))
#         return idata
#     else:
#         return loglike_dataset



# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model_pymc.pymc_io.py
"""Handling of Model Posterior Samples"""

import logging
from pathlib import Path

import arviz as az
import graphviz
from pymc.model_graph import model_to_graphviz

from . import BasePYMCModel
from ..utils.file_io import BaseFileIO

__all__ = ["PYMCIO"]

_log = logging.getLogger(__name__)


class PYMCIO(BaseFileIO):
    """Helper class to read/write NetCDF files for Arviz inference data.
    Can also write model graphs to file
    Note similar behaviour to curate.data_io.SimpleStringIO
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read_idata(
        self, mdl: BasePYMCModel = None, fn: str = "", **kwargs
    ) -> az.InferenceData:
        """Read InferenceData appropriate to a built model using
        mdl.mdl_id_fn + txtadd, or from fn"""
        txtadd = kwargs.pop("txtadd", None)
        if mdl is not None:
            fn = "_".join(filter(None, ["idata", mdl.mdl_id_fn, txtadd]))
        fqn = self.get_path_read(Path(self.snl.clean(fn)).with_suffix(".netcdf"))
        idata = az.from_netcdf(str(fqn.resolve()))
        _log.info(f"Read model idata from {str(fqn.resolve())}")
        return idata

    def write_idata(
        self, mdl: BasePYMCModel, idata: az.InferenceData = None, fn: str = "", **kwargs
    ) -> Path:
        """Accept BasePYMCModel object with attached in-sample idata, and write
        to netcdf file with name mdl.mdl_id_fn + txtadd. Optionally use this to
        write out-of-sample InferenceData passed as idata kwarg. Can implicitly
        use mdl.mdl_id_fn in either case"""
        txtadd = kwargs.pop("txtadd", None)
        if fn == "":
            fn = "_".join(filter(None, ["idata", mdl.mdl_id_fn, txtadd]))
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(".netcdf"))

        if idata is not None:
            idata.to_netcdf(str(fqn.resolve()))
        else:
            mdl.idata.to_netcdf(str(fqn.resolve()))
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn

    def write_graph(
        self,
        mdl: BasePYMCModel,
        fn: str = "",
        fmt: str = "png",
        write: bool = True,
        **kwargs,
    ) -> Path | graphviz.graphs.Digraph:
        """Accept a BasePYMCModel object mdl, get the graphviz representation
        Write to file and return the fqn to allow use within
        eda_io.FigureIO.read()
        Optionally set `write = False` and receive the graphviz directly
        """
        txtadd = kwargs.pop("txtadd", None)
        if fn == "":
            fn = "_".join(filter(None, ["graph", mdl.mdl_id_fn, txtadd]))
        fqn = self.get_path_write(f"{fn}.{fmt}")
        gv = model_to_graphviz(mdl.model, formatting="plain")
        if write == False:
            return gv
        if fmt == "png":
            gv.attr(dpi="300")
        elif fmt == "svg":
            pass
        else:
            raise ValueError('format must be in {"png", "svg"}')

        # gv auto-adds the file extension, so pre-remove if present
        gv.render(filename=str(fqn.with_suffix("")), format=fmt, cleanup=True)
        _log.info(f"Written to {str(fqn.resolve())}")
        return fqn



# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# eda.describe.py
"""Data Descriptions"""

import logging

import numpy as np
import pandas as pd
from IPython.display import display
from scipy import stats

__all__ = ["describe", "display_fw", "display_ht", "get_fts_by_dtype"]

_log = logging.getLogger(__name__)

RSD = 42
RNG = np.random.default_rng(seed=RSD)


def describe(
    df: pd.DataFrame,
    nobs: int = 3,
    nfeats: int = 30,
    limit: int = 50,  # MB
    get_mode: bool = False,
    get_counts: bool = True,
    get_cr94: bool = False,
    reset_index: bool = True,
    return_df: bool = False,
    subsample: bool = True,
    **kwargs,
) -> pd.DataFrame | None:
    """Concat transposed topN rows, numerical desc & dtypes
    Beware a dataframe full of bools or categoricals will error
    thanks to pandas.describe() being too clever
    Assume df has index.
    """

    df = df.copy()
    len_idx = df.index.nlevels
    nbytes = df.values.nbytes
    shape = df.shape
    _log.info(f"Shape: {shape}")
    _log.info(f"Memsize: {nbytes // 1e6:,.1f} MB")
    _log.info(f"Index levels: {df.index.names}")

    if nfeats + len_idx < shape[1]:
        _log.info(
            f"NOTE: nfeats + index shown {nfeats + len_idx}" + f" < width {shape[1]}"
        )

    txtadd = kwargs.pop("txtadd", None)
    limit *= 1e6
    if nbytes > limit:
        txt = (
            f"Array memsize {nbytes // 1e6:,.1f} MB >" + f" {limit / 1e6:,.1f} MB limit"
        )
        if subsample:
            df = df.sample(frac=(limit * 0.99) / nbytes, random_state=42)
            t = f"subsampled from Shape: {shape}, Memsize {nbytes / 1e6:,.1f} MB"
            txtadd = ", ".join(filter(None, [t, txtadd]))
            nbytes = df.values.nbytes
            shape = df.shape
            nobs = min(nobs, len(df))
            _log.info(txt + f", taking a subsample of {len(df)} rows")
        else:
            _log.error(txt)
            return None

    if reset_index:
        idx_new_names = [f"index: {c}" for c in list(df.index.names)]
        col_names = list(df.columns.values)
        nfeats += len(idx_new_names)
        df = df.reset_index()
        df.columns = idx_new_names + col_names

    # start with pandas describe, add on dtypes
    quantiles = [0.25, 0.5, 0.75]  # the default
    percentile_names = ["25%", "50%", "75%"]
    if get_cr94:
        quantiles = [0.03] + quantiles + [0.97]
        percentile_names = ["3%"] + percentile_names + ["97%"]
    dfdesc = df.describe(include="all", percentiles=quantiles).T

    dfout = pd.concat((dfdesc, df.dtypes), axis=1, join="outer", sort=False)
    dfout = dfout.loc[df.columns.values]
    dfout.rename(columns={0: "dtype", "unique": "count_unique"}, inplace=True)
    dfout.index.name = "ft"

    # add counts for all
    if get_counts:
        dfout["count_notnull"] = df.shape[0] - df.isnull().sum()
        dfout["count_null"] = df.isnull().sum(axis=0)
        dfout["count_inf"] = (
            np.isinf(df.select_dtypes(np.number)).sum().reindex(df.columns)
        )
        dfout["count_zero"] = (
            (df.select_dtypes(np.number) == 0).sum(axis=0).reindex(df.columns)
        )

    # add sum for numeric cols
    dfout["sum"] = np.nan
    idxs = (dfout["dtype"] == "float64") | (dfout["dtype"] == "int64")
    if np.sum(idxs.values) > 0:
        for ft in dfout.loc[idxs].index.values:
            dfout.loc[ft, "sum"] = df[ft].sum()

    # add min, max for string cols (note the not very clever overwrite of count)
    idxs = (dfout["dtype"] == "object") | (dfout["dtype"] == "string[python]")
    if np.sum(idxs.values) > 0:
        for ft in dfout.loc[idxs].index.values:
            dfout.loc[ft, "min"] = df[ft].value_counts().index.min()
            dfout.loc[ft, "max"] = df[ft].value_counts().index.max()

    fts_out_all = (
        [
            "dtype",
            "count_null",
            "count_inf",
            "count_zero",
            "count_unique",
            "top",
            "freq",
            "sum",
            "mean",
            "std",
            "min",
        ]
        + percentile_names
        + ["max"]
    )
    fts_out = [f for f in fts_out_all if f in dfout.columns.values]

    # add mode and mode count WARNING takes forever for large arrays (>10k row)
    if get_mode:
        dfnn = df.select_dtypes(exclude=np.number).copy()
        r = stats.mode(dfnn, axis=0, nan_policy="omit")
        dfmode = pd.DataFrame(
            {"mode": r[0][0], "mode_count": r[1][0]}, index=dfnn.columns
        )
        dfout = dfout.join(dfmode, how="left", left_index=True, right_index=True)
        fts_out.append(["mode", "mode_count"])

    # select summary states and prepend random rows for example cases
    rndidx = RNG.choice(np.arange(0, len(df)), size=nobs, replace=False)
    dfout = pd.concat(
        (df.iloc[rndidx].T, dfout[fts_out].copy()), axis=1, join="outer", sort=False
    )
    dfout.index.name = "ft"

    if return_df:
        return dfout
    else:
        kws_out = dict(max_rows=nfeats, shape=shape, nbytes=nbytes, txtadd=txtadd)
        display_fw(dfout.iloc[: nfeats + len_idx, :], **kws_out, **kwargs)


def display_fw(df: pd.DataFrame, **kwargs) -> None:
    """Conv fn: contextually display max cols"""

    shape = kwargs.pop("shape", df.shape)
    nbytes = kwargs.pop("nbytes", df.values.nbytes)
    txtadd = kwargs.pop("txtadd", None)

    options = {
        "display.precision": kwargs.pop("precision", 2),
        "display.max_colwidth": kwargs.pop("max_colwidth", 30),
        "display.max_rows": kwargs.pop("max_rows", 50),
        "display.max_columns": None,
    }

    if kwargs.pop("latex", False):
        options["styler.render.repr"] = "latex"
        options["styler.latex.environment"] = "longtable"

    with pd.option_context(*[i for tup in options.items() for i in tup]):
        display(df)
        t = f"Shape: {shape}, Memsize {nbytes / 1e6:,.1f} MB"
        display(", ".join(filter(None, [t, txtadd])))


def display_ht(df: pd.DataFrame, nrows=3, **kwargs) -> None:
    """Convenience fn: Display head and tail n rows via display_fw"""

    nrows = min(nrows, len(df))
    dfd = df.iloc[np.r_[0:nrows, -nrows:0]].copy()
    display_fw(dfd, shape=df.shape, nbytes=df.values.nbytes, **kwargs)


def get_fts_by_dtype(df: pd.DataFrame, as_dataframe: bool = False) -> dict:
    """Return a dictionary of lists of feats within df according to dtype"""
    dtypes = df.dtypes.to_dict().items()
    fts = dict(
        categorical=[k for k, v in dtypes if v.name[:3] == "cat"],  # category
        cat=[
            k
            for k, v in df.dtypes.to_dict().items()
            if (v.name[:3] == "obj") | (v.name[:3] == "str")
        ],
        bool=[k for k, v in dtypes if v.name == "bool"],
        boolean=[k for k, v in dtypes if v.name == "boolean"],
        datetime=[k for k, v in dtypes if v.name[:3] == "dat"],
        int=[k for k, v in dtypes if v.name[:3] == "int"],
        float=[k for k, v in dtypes if v.name[:3] == "flo"],
    )
    w = []
    for _, v in fts.items():
        w += v

    mismatch = list(set(df.columns) - set(w))
    if len(mismatch) > 0:
        raise ValueError(
            f"Failed to match a dtype to {mismatch}"
            + f"\nThese fts did match correctly: {fts}"
        )

    if as_dataframe:
        dtypes = ["categorical", "cat", "bool", "boolean", "datetime", "int", "float"]
        d = {w: k for k, v in fts.items() for w in v}
        dfd = pd.DataFrame.from_dict(d, orient="index", columns=["dtype"])
        dfd.index.set_names("ft", inplace=True)
        dfd["dtype"] = pd.Categorical(dfd["dtype"], categories=dtypes, ordered=True)
        return dfd

    return fts



# Copyright 2026 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# eda.calc.py
"""Calculations to help EDA"""

import logging
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import figure
from scipy import stats
from sklearn.decomposition import TruncatedSVD
from sklearn.preprocessing import StandardScaler

RSD = 42
rng = np.random.default_rng(seed=RSD)

_log = logging.getLogger(__name__)

__all__ = [
    "fit_and_plot_fn",
    "get_gini",
    "bootstrap_index_only",
    "bootstrap",
    "bootstrap_lr",
    "calc_geometric_cv",
    "calc_location_in_ecdf",
    "month_diff",
    "tril_nan",
    "calc_svd",
]


# TODO see issue #2
def fit_and_plot_fn(obs: pd.Series) -> tuple[figure.Figure, dict]:
    """Fit dists to 1d array of `obs`, report RMSE and plot the fits
    see https://stackoverflow.com/a/37616966
    """

    dists_discrete_support_gte_zero = {
        "poisson": stats.poisson
        # 'binom': stats.binom,  # placeholder
        # 'nbinom': stats.nbinom,  # placeholder
    }

    dists_cont_support_real = {
        "norm": stats.norm,
        "cauchy": stats.cauchy,
        "gumbel": stats.gumbel_r,
        "invweibull": stats.invweibull,
    }

    dists_cont_support_gte_zero = {
        "expon": stats.expon,
        "halfnorm": stats.halfnorm,
        "halfcauchy": stats.halfcauchy,
        "fisk": stats.fisk,
    }

    dists_cont_support_gt_zero = {
        "gamma": stats.gamma,
        "invgamma": stats.invgamma,
        "lognorm": stats.lognorm,
    }

    nbins = 50
    dist_kind = "Continuous"
    params = {}
    f, ax1d = plt.subplots(1, 1, figsize=(14, 6))
    hist_kws = dict(
        kde=False, label="data", ax=ax1d, alpha=0.5, color="#aaaaaa", zorder=-1
    )
    line_kws = dict(lw=2, ls="--", ax=ax1d)

    # handle nans and infs
    n_nans = pd.isnull(obs).sum()
    n_infs = np.isinf(obs).sum()
    idx = pd.isnull(obs) | np.isinf(obs)
    obs = obs.loc[~idx]

    obs_is_discrete = sum(obs == (obs // 1)) == len(obs)
    n_zeros = (obs == 0).sum()
    n_negs = (obs < 0).sum()

    def _annotate_facets(n_nans, n_infs, n_zeros):
        """Convenience to annotate, based on eda.plots.plot_float_dist"""
        n_zeros = (obs == 0).sum()
        mean = obs.mean()
        med = obs.median()
        ax = plt.gca()
        ax.text(
            0.5,
            0.97,
            (
                f"NaNs: {n_nans},  infs+/-: {n_infs},  zeros: {n_zeros},  "
                + f"mean: {mean:.2f},  med: {med:.2f}"
            ),
            transform=ax.transAxes,
            ha="center",
            va="top",
            backgroundcolor="w",
            fontsize=10,
        )

    if obs_is_discrete:
        dist_kind = "Discrete"
        nbins = np.int(obs.max())
        obs_count, bin_edges = np.histogram(obs, bins=nbins, density=False)
        bin_centers = (bin_edges + np.roll(bin_edges, -1))[:-1] / 2.0
        bin_centers_int = np.round(bin_centers)
        dists = dists_discrete_support_gte_zero

        _ = sns.histplot(x=obs, bins=nbins, stat="count", **hist_kws)

        # TODO fix this hack: only works for poisson right now
        for d, _dist in dists.items():
            mle_pois = obs.mean()
            params[d] = dict(shape=mle_pois, loc=0, scale=None)
            pmf = stats.poisson.pmf(k=bin_centers_int, mu=mle_pois) * len(obs)

            # rmse not necessarily good for discrete count models
            # https://stats.stackexchange.com/questions/48811/cost-function-for-validating-poisson-regression-models
            rmse = np.sqrt(np.sum(np.power(obs_count - pmf, 2.0)) / len(obs))
            _ = sns.lineplot(
                x=bin_centers_int, y=pmf, label=f"{d}: {rmse:#.2g}", **line_kws
            )

    else:
        obs_density, bin_edges = np.histogram(obs, bins=nbins, density=True)
        bin_centers = (bin_edges + np.roll(bin_edges, -1))[:-1] / 2.0

        dists = dists_cont_support_real
        if n_negs == 0:
            dists.update(dists_cont_support_gte_zero)
        if n_zeros == 0:
            dists.update(dists_cont_support_gt_zero)

        _ = sns.histplot(x=obs, bins=nbins, stat="density", **hist_kws)

        for d, dist in dists.items():
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore")
                ps = dist.fit(obs, floc=0)  # can throw RuntimeWarnings which we ignore
            shape, loc, scale = ps[:-2], ps[-2], ps[-1]
            params[d] = dict(shape=shape, loc=loc, scale=scale)
            pdf = dist.pdf(bin_centers, loc=loc, scale=scale, *shape)
            rmse = np.sqrt(np.sum(np.power(obs_density - pdf, 2.0)) / len(obs))
            _ = sns.lineplot(
                x=bin_centers, y=pdf, label=f"{d}: {rmse:#.2g}", **line_kws
            )

    title = f"{dist_kind} function approximations to {obs.name}"
    _ = f.suptitle(title, y=0.97)
    _ = f.axes[0].legend(title="dist: RMSE", title_fontsize=10)
    _annotate_facets(n_nans, n_infs, n_zeros)

    return f, params


def get_gini(r: np.ndarray, n: np.ndarray) -> np.ndarray:
    """For array r, return estimate of gini co-efficient over n
    g = A / (A+B)
    """
    return 1 - sum(r.sort_values().cumsum() * (2 / n))


def bootstrap_index_only(a: np.ndarray, nboot: int = None) -> np.ndarray:
    """Create a 2D index for vectorised resampling via numpy broadcasting
    If nboot is None, use len(a): this preserves sample error in the resamples,
    and allows the resample to behave with the same variance as the sample a.
    """
    if not isinstance(a, np.ndarray):
        raise ValueError("must supply a as np.array to broadcast 2D properly")
    rng = np.random.default_rng(seed=RSD)
    if nboot is None:
        nboot = len(a)
    return rng.integers(0, len(a), size=(len(a), nboot))


def bootstrap(a: np.ndarray, nboot: int = None, summary_fn=None) -> np.ndarray:
    """Calc vectorised bootstrap resample of ndarray of observations with
    optional summary_fn which must be the numpy function to be used in
    np.apply_along_axis. e.g. np.mean or np.sum
    NOTE nboot will default to len(a) see https://sedar.co/posts/bootstrap-primer/
    """
    sample_idx = bootstrap_index_only(a, nboot)
    samples = a[sample_idx]
    if summary_fn is not None:
        return np.apply_along_axis(summary_fn, 0, samples)
    else:
        return samples


def bootstrap_lr(
    df: pd.DataFrame, prm: str = "premium", clm: str = "claim", nboot: int = 1000
) -> pd.DataFrame:
    """Calc vectorised bootstrap loss ratios for df one-row-per-policy
    Pass dataframe or group, accept nans in clm
    Use the same index for prem and claims
    """
    idx = bootstrap_index_only(a=df[prm].values, nboot=nboot)
    s_prm = df[prm].values[idx]
    s_clm = df[clm].values[idx]

    dfboot = pd.DataFrame(
        {
            "premium_sum": np.apply_along_axis(np.sum, 0, s_prm),
            "claim_sum": np.apply_along_axis(np.sum, 0, np.nan_to_num(s_clm, 0)),
        }
    )

    dfboot["lr"] = dfboot["claim_sum"] / dfboot["premium_sum"]
    return dfboot


def calc_geometric_cv(lognormal_yhat: np.ndarray) -> np.ndarray:
    """Calculate geometric coefficient of variation for log-normally
    distributed samples.
    Expect 2D array shape (nobs, nsamples)
    https://en.wikipedia.org/wiki/Coefficient_of_variation#Log-normal_data
    """
    return np.sqrt(np.exp(np.std(np.log(lognormal_yhat), axis=1) ** 2) - 1)


def _ecdf(a: np.ndarray) -> tuple:
    """Empirical CDF of array
    Return sorted array and ecdf values
    """
    x = np.sort(a)
    n = len(a)
    y = np.arange(1, n + 1) / n
    return x, y


def calc_location_in_ecdf(baseline_arr, test_arr):
    """Calculate the position of each element in test_arr
    relative to the ECDF described by baseline_arr
    (len(baseline_arr) === len(test_arr)) = False
    """
    sorted_baseline, cdf_prop = _ecdf(baseline_arr)
    idxs = np.argmax((test_arr < sorted_baseline[:, None]), axis=0)
    return cdf_prop[idxs]


def month_diff(
    a: pd.Series, b: pd.Series, series_name: str = "month_diff"
) -> pd.Series:
    """Integer month count between dates a to b
    https://stackoverflow.com/a/40924041/1165112
    In recent pandas can equally use to_period(), though it's unwieldy
    e.g
    [x.n for x in (df['obs_date'].dt.to_period('M') -
                    df['reported_date'].dt.to_period('M'))]
    """
    s = 12 * (b.dt.year - a.dt.year) + (b.dt.month - a.dt.month)
    s = s.astype(int)
    s.name = series_name
    return s


def tril_nan(m: np.ndarray, k: int = 0) -> np.ndarray:
    """Copy of np.tril but mask with np.nans not zeros
    Example usage, to set tril to np.nan in DataFrame.corr()
    corr = df_b0.corr()
    mask = eda.tril_nan(corr, k=-1)
    corr.mask(np.isnan(mask))

    """

    m = np.asanyarray(m)  # numpy.core.numeric
    mask = np.tri(*m.shape[-2:], k=k, dtype=bool)

    # return np.where(mask, m, np.ones(1, m.dtype) * np.nan)
    return np.where(mask, m, np.nan)


def calc_svd(df: pd.DataFrame, k: int = 10) -> tuple[pd.DataFrame, TruncatedSVD]:
    """Calc SVD for k components (and preprocess to remove nulls and zscore),
    report degeneracy, return transformed df and fitted TruncatedSVD object"""

    # protect SVD from nulls
    idx_nulls = df.isnull().sum(axis=1) > 0
    if sum(idx_nulls) > 0:
        df = df.loc[~idx_nulls].copy()
        _log.info(f"Excluding {sum(idx_nulls)} rows containing a null, prior to SVD")

    # standardize
    scaler = StandardScaler().fit(df)
    dfs = pd.DataFrame(scaler.transform(df), index=df.index, columns=df.columns)

    # use scikit learn's TruncatedSVD with randomized
    k = min(k, dfs.shape[1] - 1)
    svd = TruncatedSVD(n_components=k, random_state=RSD)
    svd_fit = svd.fit(dfs)

    # Are any eigenvalues NaN or really small?
    n_null = sum(np.isnan(svd_fit.singular_values_))
    assert n_null == 0, f"{n_null} Singular Values are NaN"
    n_tiny = sum(svd_fit.singular_values_ < 1e-12)
    assert n_tiny == 0, f"{n_tiny} Singular Values are < 1e-12"

    dfx = svd_fit.transform(dfs)

    return dfx, svd_fit
